"""
This file contains constants that are used to manage OCP Image and RPM builds
"""

# config data pulled from here
OCP_BUILD_DATA_RO = "https://github.com/openshift/ocp-build-data"
# above is used so that anyone can clone
OCP_BUILD_DATA_RW = "git@github.com:openshift/ocp-build-data.git"

BREW_HUB = "https://brewhub.engineering.redhat.com/brewhub"
BREW_IMAGE_HOST = "brew-pulp-docker01.web.prod.ext.phx2.redhat.com:8888"
CGIT_URL = "http://pkgs.devel.redhat.com/cgit"

# For Bugzilla searches
BUGZILLA_SERVER = "bugzilla.redhat.com"
DEFAULT_VERSIONS = (
    "3.0.0",
    "3.1.0", "3.1.1",
    "3.2.0", "3.2.1",
    "3.3.0", "3.3.1",
    "3.4.0", "3.4.1",
    "3.5.0", "3.5.1",
    "3.6.0", "3.6.1",
    "3.7.0", "3.7.1",
    "3.8.0", "3.8.1",
    "3.9.0", "3.9.1",
    "3.10.0", "3.10.1",
    "3.11.0", "3.11.1",
    "unspecified"
)
VALID_BUG_STATES = ['NEW', 'ASSIGNED', 'POST', 'MODIFED', 'ON_QA', 'VERIFIED', 'RELEASE_PENDING', 'CLOSED']

errata_url = "https://errata.devel.redhat.com"
bugzilla_query_url = 'https://bugzilla.redhat.com/buglist.cgi?bug_status=MODIFIED&classification=Red%20Hat&f1=component&f2=component&f3=component&f4=cf_verified&keywords=UpcomingRelease&keywords_type=nowords&known_name=All%203.x%20MODIFIED%20Bugs&list_id=8111122&o1=notequals&o2=notequals&o3=notequals&o4=notequals&product=OpenShift%20Container%20Platform&query_format=advanced&short_desc=%5C%5Bfork%5C%5D&short_desc_type=notregexp&{0}&v1=RFE&v2=Documentation&v3=Security&v4=FailedQA&version=3.0.0&version=3.1.0&version=3.1.1&version=3.2.0&version=3.2.1&version=3.3.0&version=3.3.1&version=3.4.0&version=3.4.1&version=3.5.0&version=3.5.1&version=3.6.0&version=3.6.1&version=3.7.0&version=3.7.1&version=3.8.0&version=3.8.1&version=3.9.0&version=3.9.1&version=3.10.0&version=3.10.1&version=3.11.0&version=3.11.1&version=unspecified'

# For new errata
errata_solution = """Before applying this update, make sure all previously released errata relevant to your system have been applied.

For OpenShift Container Platform 3.{Y} see the following documentation, which will be updated shortly for release 3.{Y}.z, for important instructions on how to upgrade your cluster and fully apply this asynchronous errata update:

https://docs.openshift.com/container-platform/3.{Y}/release_notes/ocp_3_{Y}_release_notes.html

This update is available via the Red Hat Network. Details on how to use the Red Hat Network to apply this update are available at https://access.redhat.com/articles/11258."""
errata_description = """Red Hat OpenShift Container Platform is Red Hat's cloud computing Kubernetes application platform solution designed for on-premise or private cloud deployments.

This advisory contains the RPM packages for Red Hat OpenShift Container Platform 3.{Y}.z. See the following advisory for the container images for this release:

https://access.redhat.com/errata/RHBA-2018:0114

Space precludes documenting all of the bug fixes and enhancements in this advisory. See the following Release Notes documentation, which will be updated shortly for this release, for details about these changes:

https://docs.openshift.com/container-platform/3.{Y}/release_notes/ocp_3_{Y}_release_notes.html

All OpenShift Container Platform 3.{Y} users are advised to upgrade to these updated packages and images."""

# QE Group name
errata_quality_responsibility_name = 'OpenShift QE'
errata_synopsis = {
    'rpm': 'OpenShift Container Platform 3.{Y} bug fix and enhancement update',
    'image': 'OpenShift Container Platform 3.{Y} images update'
}
errata_topic = 'Red Hat OpenShift Container Platform releases 3.{Y}.z are now available with updates to packages and images that fix several bugs and add enhancements.'
# 1965 => (RHBA; Active; Product: RHOSE; Devel Group: ENG OpenShift
#          Enterprise; sorted by newest)
# https://errata.devel.redhat.com/filter/1965
errata_default_filter = '1965'
# 1991 => (Active; Product: RHOSE; Devel Group: ENG OpenShift
#          Enterprise; sorted by newest)
# https://errata.devel.redhat.com/filter/1991
errata_live_advisory_filter = '1991'
# 2051 => (RHBA; State REL PREP, PUSH READY, IN PUSH, SHIPPED;
#          Product: RHOSE; Devel Group: ENG OpenShift Enterprise;
#          sorted by newest)
# https://errata.devel.redhat.com/filter/2051
errata_immutable_advisory_filter = '2051'
errata_active_advisory_labels = [
    "NEW_FILES",
    "QE",
    "REL_PREP",
    "PUSH_READY",
    "IN_PUSH"
]
errata_inactive_advisory_labels = [
    "SHIPPED_LIVE",
    "DROPPED_NO_SHIP"
]
errata_valid_impetus = [
    'standard',
    'cve',
    'ga',
    'test'
]
######################################################################
# API endpoints with string formatting placeholders as
# necessary. Index of all available endpoints is available in the
# online documentation.
#
# https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-index-by-url
errata_add_bug_url = errata_url + "/api/v1/erratum/{id}/add_bug"
errata_add_build_url = errata_url + "/api/v1/erratum/{id}/add_build"
errata_add_builds_url = errata_url + "/api/v1/erratum/{id}/add_builds"
errata_add_comment_url = errata_url + "/api/v1/erratum/{id}/add_comment"
errata_bug_refresh_url = errata_url + "/api/v1/bug/refresh"
errata_change_state_url = errata_url + "/api/v1/erratum/{id}/change_state"
errata_filter_list_url = errata_url + "/filter/{id}.json"
errata_get_build_url = errata_url + "/api/v1/build/{id}"
errata_get_builds_url = errata_url + "/api/v1/erratum/{id}/builds"
errata_get_comment_url = errata_url + "/api/v1/comments/{id}"
errata_get_comments_url = errata_url + "/api/v1/comments"
errata_get_erratum_url = errata_url + "/api/v1/erratum/{id}"
errata_post_erratum_url = errata_url + "/api/v1/erratum"
######################################################################
# Scaffolding for creating a new advisory. See the online
# documentation for a description of all allowed fields, including
# which are required and which are optional
#
# https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratum
errata_new_object = {
    'product': 'RHOSE',
    'release': 'RHOSE ASYNC',
    'advisory': {
        'assigned_to_email': None,  # Provided as input
        'description': errata_description,
        'errata_type': 'RHBA',
        'idsfixed': '',
        'manager_email': None,  # Provided as input
        'package_owner_email': None,  # Provided as input
        'publish_date_override': None,
        'quality_responsibility_name': errata_quality_responsibility_name,
        'security_impact': 'None',
        'solution': errata_solution,
        'synopsis': None,
        'topic': errata_topic,
    }
}
######################################################################
# We ship updates for various architectures. Images are
# architecture-agnostic
errata_shipped_arches = {
    'rpm': ['--arch=x86_64', '--arch=noarch'],
    'image': [''],
}
# Do not include these in tag results (carried over legacy behavior
# from update-errata-with-diffs.sh
errata_package_excludes = ["cockpit", "atomic-openshift-clients-redistributable",
                           "python-jsonschema", "python-ruamel-yaml", "python-wheel",
                           "python-typing", "python-ruamel-ordereddict", "ansible"]

bugzilla_invalid_transition_comment = """There is a potential issue with this bug that may prevent it from being processed by our automation.

For more information on proper bug management visit:
https://mojo.redhat.com/docs/DOC-1178565#jive_content_id_How_do_I_get_my_Bugzilla_Bug_to_VERIFIED
"""

DISTGIT_MAX_FILESIZE = 50000000



"""Utility functions for general interactions with the errata SERVICE.

As a human, you will be working with "ADVISORIES". Advisories are one
or more errata (an errata is one or more erratum) as well as
associated metadata.

Classes representing an ERRATUM (a single errata)

"""

import copy
import datetime
import json

import constants
import brew
import exceptions

import requests
from requests_kerberos import HTTPKerberosAuth


def get_erratum(id):
    """5.2.1.2. GET /api/v1/erratum/{id}

    Retrieve the advisory data.

    https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-get-apiv1erratumid

    :return SUCCESS: An Erratum object
    :return FAILURE: :bool:False
    :raises: exceptions.ErrataToolUnauthenticatedException if the user is not authenticated to make the request
    """
    res = requests.get(constants.errata_get_erratum_url.format(id=id),
                       auth=HTTPKerberosAuth())

    if res.status_code == 200:
        return Erratum(body=res.json())
    elif res.status_code == 401:
        raise exceptions.ErrataToolUnauthenticatedException(res.text)
    else:
        return False


def find_mutable_erratum(kind, minor, major=3):
    """Find the latest mutable (may be updated/changed) erratum for a
    release series using a combination of search parameters.

    :param string kind: One of 'rpm' or 'image'
    :param str/int minor: The minor release to search against (i.e.,
        3.10, minor=10)
    :param str/int major: [Default 3] The major release to search
        against (i.e., 3.10, major=3)

    Example:

    To find a 3.10 erratum that you could attach new image builds to:
    * kind='image', minor=10

    Return Values:

    If an open AND mutable erratum exists matching your search:
    :return: An `Erratum` object of the discovered erratum

    If no open OR mutable erratum exists matching your search:
    :return: `None`

    DEV NOTE: If this function returns `None` then you can use the
    other find_*_erratum function to find the latest erratum of kind
    `kind` in the given release series. That is to say, if there are
    no open erratum for your release series you can still find the
    most recent erratum, even if it has passed the `QA` state or has
    even already been released.

    :raises: exceptions.ErrataToolUnauthorizedException if the user is not authenticated to make the request

    """
    pass


def find_latest_erratum(kind, minor, major=3):
    """Find an erratum in a given release series, in ANY state.

    Put simply, this tells you the erratum that has the most recent,
    or furthest in the future, release date set.

    This is useful for determining the release date of a new
    erratum. This combines the functionality provided by
    find_mutable_erratum and extends it by including searching closed
    or immutable erratum. These two functions can work well in tandem.

    Contrast this with find_mutable_erratum (best suited for elliott
    actions explicitly designed to UPDATE an erratum), this function
    promises to tell you the freshest release date of erratum in any
    state in a given release series.

    Example:

    You are creating a new erratum. The latest erratum in that series
    MAY or MAY NOT have gone to SHIPPED_LIVE state yet. Regardless of
    that, this function will tell you what the latest ship date is for
    an erratum in that series.

    If erratum exists matching your search:
    :return: An `Erratum` object of the erratum

    If no erratum can be found matching your search:
    :return: `None`
    """
    release = "{}.{}".format(major, minor)
    found_advisory = None

    # List of hashes because we will scan the Mutable advisories first
    filters = [
        {'Mutable Advisories': constants.errata_default_filter},
        {'Immutable Advisories': constants.errata_immutable_advisory_filter}
    ]

    # Fetch initial lists of advisories in each pre-defined filter
    advisory_list = []
    print("Running initial advisory fetching")
    for f in filters:
        state_desc, filter_id = f.items()[0]
        print("Fetching {state}".format(state=state_desc))
        advisories = get_filtered_list(filter_id, limit=50)
        # Filter out advisories that aren't for this release
        advisory_list.extend([advs for advs in advisories if " {} ".format(release) in advs.synopsis])
        print("Advisory list has {n} items after this fetch".format(
            n=len(advisory_list)))

    print("Looking for elliott metadata in comments:")
    matched_advisories = []
    for advisory in advisory_list:
        print("Scanning advisory {}".format(str(advisory)))

        for c in advisory.get_comments():
            try:
                metadata = json.loads(c['attributes']['text'])
            except Exception as e:
                pass
            else:
                if str(metadata['release']) == str(release) and metadata['kind'] == kind and metadata['impetus'] == 'standard':
                    matched_advisories.append(advisory)
                    # Don't scan any more comments
                    break

    if matched_advisories == []:
        return None
    else:
        # loop over discovered advisories, select one with max() date
        real_advisories = [get_erratum(e.advisory_id) for e in matched_advisories]
        sorted_dates = sorted(real_advisories, key=lambda advs: advs.release_date)
        return sorted_dates[-1]


def new_erratum(kind=None, release_date=None, create=False, minor='Y',
                assigned_to=None, manager=None, package_owner=None):
    """5.2.1.1. POST /api/v1/erratum

    Create a new advisory.
    Takes an unrealized advisory object and related attributes using the following format:

    https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratum

    :param string kind: One of 'rpm' or 'image', effects boilerplate text
    :param string release_date: A date in the form YYYY-MM-DD
    :param bool create: If true, create the erratum in the Errata
        tool, by default just the DATA we would have POSTed is
        returned
    :param str/int minor: The minor release to substitute into the
        errata boilerplate text (see:
        :mod:`constants`). E.g., if this is a '3.9'
        release, you would provide '9' as the value for 'minor'
    :param string assigned_to: The email address of the group responsible for
        examining and approving the advisory entries
    :param string manager: The email address of the manager responsible for
        managing the contents and status of this advisory
    :param string package_owner: The email address of the person who is handling
        the details and status of this advisory

    :return: An Erratum object
    :raises: exceptions.ErrataToolUnauthenticatedException if the user is not authenticated to make the request
    """
    if release_date is None:
        release_date = datetime.datetime.now() + datetime.timedelta(days=21)

    if kind is None:
        kind = 'rpm'

    body = copy.deepcopy(constants.errata_new_object)
    body['advisory']['publish_date_override'] = release_date
    body['advisory']['synopsis'] = constants.errata_synopsis[kind].format(Y=minor)

    # Fill in the minor version variables
    body['advisory']['description'] = body['advisory']['description'].format(Y=minor)
    body['advisory']['solution'] = body['advisory']['solution'].format(Y=minor)
    body['advisory']['synopsis'] = body['advisory']['synopsis'].format(Y=minor)
    body['advisory']['topic'] = body['advisory']['topic'].format(Y=minor)

    body['advisory']['assigned_to_email'] = assigned_to
    body['advisory']['manager_email'] = manager
    body['advisory']['package_owner_email'] = package_owner

    if create:
        # THIS IS NOT A DRILL
        res = requests.post(constants.errata_post_erratum_url,
                            auth=HTTPKerberosAuth(),
                            json=body)

        if res.status_code == 201:
            return Erratum(body=res.json())
        elif res.status_code == 401:
            raise exceptions.ErrataToolUnauthenticatedException(res.text)
        else:
            raise exceptions.ErrataToolError("Other error (status_code={code}): {msg}".format(
                code=res.status_code,
                msg=res.text))
    else:
        # This is a noop
        return json.dumps(body, indent=2)


def get_filtered_list(filter_id=constants.errata_default_filter, limit=5):
    """return a list of Erratum() objects from results using the provided
filter_id

    :param filter_id: The ID number of the pre-defined filter
    :param int limit: How many erratum to list
    :return: A list of Erratum objects

    :raises exceptions.ErrataToolUnauthenticatedException: If the user is not authenticated to make the request
    :raises exceptions.ErrataToolError: If the given filter does not exist, and, any other unexpected error

    Note: Errata filters are defined in the ET web interface
    """
    filter_endpoint = constants.errata_filter_list_url.format(
        id=filter_id)
    res = requests.get(filter_endpoint,
                       auth=HTTPKerberosAuth())
    if res.status_code == 200:
        # When asked for an advisory list which does not exist
        # normally you would expect a code like '404' (not
        # found). However, the Errata Tool sadistically returns a 200
        # response code. That leaves us with one option: Decide that
        # successfully parsing the response as a JSONinfo object indicates
        # a successful API call.
        try:
            return [Erratum(body=advs) for advs in res.json()][:limit]
        except Exception:
            raise exceptions.ErrataToolError("Could not locate the given advisory filter: {fid}".format(
                fid=filter_id))
    elif res.status_code == 401:
        raise exceptions.ErrataToolUnauthenticatedException(res.text)
    else:
        raise exceptions.ErrataToolError("Other error (status_code={code}): {msg}".format(
            code=res.status_code,
            msg=res.text))


class Erratum(object):
    """
    Model for interacting with individual Erratum. Erratum instances
    can be created from the brief info provided in a filtered list, as
    well as from the full erratum body returned by the Errata Tool
    API. See also: get_erratum(id) for creating a filled in Erratum
    object automatically.
    """

    date_format = '%Y-%m-%dT%H:%M:%SZ'

    def __init__(self, body=None):
        """If a `body` is provided then this is an EXISTING advisory and we
        are filling in all the  at the time of object creation.
        """
        self.body = body

        if body is not None:
            self._parse_body()
        else:
            # Attributes used in str() representation
            self.adv_name = ''
            self.synopsis = ''
            self.url = ''
            self.created_at = datetime.datetime.now()
            self.release_date = None

    ######################################################################
    # Basic utility/setup methods

    def __str__(self):
        return "{date} {state} {synopsis} {url}".format(
            date=self.created_at.isoformat(),
            state=self.status,
            synopsis=self.synopsis,
            url=self.url)

    def _parse_body(self):
        """The `body` content is different based on where it came from."""
        # Erratum from the direct erratum GET method
        if 'params' in self.body:
            # An advisory will come back as one of the following
            # kinda, we're just not sure which until we read the
            # object body
            for kind in ['rhba', 'rhsa', 'rhea']:
                if kind in self.body['errata']:
                    # Call it just a generic "red hat advisory"
                    rha = self.body['errata'][kind]
                    break

            content = self.body['content']['content']

            self.advisory_id = rha['id']
            self.advisory_name = rha['fulladvisory']
            self.synopsis = rha['synopsis']
            self.description = content['description']
            self.solution = content['solution']
            self.topic = content['topic']
            self.status = rha['status']
            self.created_at = datetime.datetime.strptime(rha['created_at'], self.date_format)
            self.release_date = datetime.datetime.strptime(rha['publish_date_override'], self.date_format)
            self.url = "{et}/advisory/{id}".format(
                et=constants.errata_url,
                id=self.advisory_id)
        else:
            # Erratum returned from an advisory filtered list
            self.advisory_id = self.body.get('id', 0)
            self.advisory_name = self.body.get('advisory_name', '')
            self.synopsis = self.body['synopsis']
            self.description = self.body['content']['description']
            self.solution = self.body['content']['solution']
            self.topic = self.body['content']['topic']
            self.status = self.body.get('status', 'NEW_FILES')
            self.created_at = datetime.datetime.strptime(self.body['timestamps']['created_at'], self.date_format)
            self.url = "{et}/advisory/{id}".format(
                et=constants.errata_url,
                id=self.advisory_id)

    def refresh(self):
        """Refreshes this object by pulling down a fresh copy from the API"""
        self.body = get_erratum(self.advisory_id).body
        self._parse_body()

    def to_json(self):
        return json.dumps(self.body, indent=2)

    ######################################################################
    # The following methods are related to REST API interactions

    def add_bugs(self, bugs=[]):  # pragma: no cover
        """Shortcut for several calls to self.add_bug()

        :param Bug bugs: A list of :module:`bugzilla` Bug objects
        """
        for bug in bugs:
            yield (self.add_bug(bug), bug.id)

    def add_bug(self, bug):
        """5.2.1.5. POST /api/v1/erratum/{id}/add_bug

        Add a bug to an advisory.

        Example request body:
            {"bug": "884202"}

        https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratumidadd_bug

        :param Bug bug: A :module:`bugzilla` Bug object
        """
        return requests.post(constants.errata_add_bug_url.format(id=self.advisory_id),
                             auth=HTTPKerberosAuth(),
                             json={'bug': bug.id})

    def add_builds(self, builds=[]):
        """5.2.2.7. POST /api/v1/erratum/{id}/add_builds

        Add one or more brew builds to an advisory.

        Example request body:
            {"product_version": "RHEL-7", "build": "rhel-server-docker-7.0-23", "file_types": ["ks","tar"]}

        The request body is a single object or an array of objects
        specifying builds to add, along with the desired product
        version (or pdc release) and file type(s). Builds may be
        specified by ID or NVR.

        https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratumidadd_builds

        :param list[Build] builds: List of Build objects to attach to
        an advisory

        :return: True if builds were added successfully

        :raises: exceptions.BrewBuildException if the builds could not be attached
        :raises: exceptions.ErrataToolUnauthenticatedException if the user is not authenticated to make the request
        """
        data = [b.to_json() for b in builds]

        res = requests.post(constants.errata_add_builds_url.format(id=self.advisory_id),
                            auth=HTTPKerberosAuth(),
                            json=data)

        print(res.status_code)
        print(res.text)

        if res.status_code == 422:
            # "Something" bad happened
            print(res.status_code)
            print(res.text)
            raise exceptions.BrewBuildException(str(res.json()))
        elif res.status_code == 401:
            raise exceptions.ErrataToolUnauthenticatedException(res.text)
        # TODO: Find the success return code
        else:
            return True

    def add_comment(self, comment):  # pragma: no cover
        """5.2.1.8. POST /api/v1/erratum/{id}/add_comment

        Add a comment to an advisory.
        Example request body:

            {"comment": "This is my comment"}

        The response body is the updated or unmodified advisory, in the same format as GET /api/v1/erratum/{id}.

        https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratumidadd_comment

        :param dict comment: The metadata object to add as a comment
        """
        data = {"comment": json.dumps(comment)}
        return requests.post(constants.errata_add_comment_url.format(id=self.advisory_id),
                             auth=HTTPKerberosAuth(),
                             data=data)

    def change_state(self, state):
        """5.2.1.14. POST /api/v1/erratum/{id}/change_state

        Change the state of an advisory.
        Example request body:

            {"new_state": "QE"}

        Request body may contain:
            new_state: e.g. 'QE' (required)
            comment: a comment to post on the advisory (optional)

        :param str state: The state to change the advisory to
        :return: True on successful state change
        :raises: exceptions.ErrataToolUnauthenticatedException if the user is not authenticated to make the request

        https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-post-apiv1erratumidchange_state
        """
        res = requests.post(constants.errata_change_state_url.format(id=self.advisory_id),
                            auth=HTTPKerberosAuth(),
                            data={"new_state": state})

        # You may receive this response when: Erratum isn't ready to
        # move to QE, no builds in erratum, erratum has no Bugzilla
        # bugs or JIRA issues.
        if res.status_code == 422:
            # Conditions not met
            raise exceptions.ErrataToolError("Can not change erratum state, preconditions not yet met. Error message: {msg}".format(
                msg=res.text))
        elif res.status_code == 401:
            raise exceptions.ErrataToolUnauthenticatedException(res.text)
        elif res.status_code == 201:
            # POST processed successfully
            self.refresh()
            return True

    def get_comments(self):
        """5.2.10.2. GET /api/v1/comments?filter[key]=value

        Retrieve all advisory comments
        Example request body:

            {"filter": {"errata_id": 11112, "type": "AutomatedComment"}}

        Returns an array of comments ordered in descending order
        (newest first). The array may be empty depending on the filters
        used. The meaning of each attribute is documented under GET
        /api/v1/comments/{id} (see Erratum.get_comment())

        Included for reference:
        5.2.10.2.1. Filtering

        The list of comments can be filtered by applying
        filter[key]=value as a query parameter. All attributes of a
        comment - except advisory_state - can be used as a filter.

        This is a paginated API. Reference documentation:
        https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-pagination
        """
        body = {
            "filter": {
                "errata_id": self.advisory_id,
                "type": "Comment"
                }
            }
        res = requests.get(constants.errata_get_comments_url,
                           auth=HTTPKerberosAuth(),
                           json=body)

        if res.status_code == 200:
            return res.json().get('data', [])
        elif res.status_code == 401:
            raise exceptions.ErrataToolUnauthorizedException(res.text)
        else:
            return False



"""
Utility functions for general interactions with Brew and Builds
"""

# stdlib
import ast
import time
import datetime
import subprocess
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing import cpu_count
from multiprocessing import Lock
import shlex
import koji
import koji_cli.lib
import traceback

# ours
import constants
import exceptions
import exectools
import logutil

# 3rd party
import click
import requests
from requests_kerberos import HTTPKerberosAuth

logger = logutil.getLogger(__name__)

# ============================================================================
# Brew/Koji service interaction functions
# ============================================================================

# Populated by watch_task. Each task_id will be a key in the dict and
# each value will be a TaskInfo: https://github.com/openshift/enterprise-images/pull/178#discussion_r173812940
watch_task_info = {}
# Protects threaded access to watch_task_info
watch_task_lock = Lock()


def get_watch_task_info_copy():
    """
    :return: Returns a copy of the watch_task info dict in a thread safe way. Each key in this dict
     is a task_id and each value is a koji TaskInfo with potentially useful data.
     https://github.com/openshift/enterprise-images/pull/178#discussion_r173812940
    """
    with watch_task_lock:
        return dict(watch_task_info)


def watch_task(log_f, task_id, terminate_event):
    end = time.time() + 4 * 60 * 60
    watcher = koji_cli.lib.TaskWatcher(
        task_id,
        koji.ClientSession(constants.BREW_HUB, opts={'serverca': '/etc/pki/brew/legacy.crt'}),
        quiet=True)
    error = None
    except_count = 0
    while error is None:
        try:
            watcher.update()
            except_count = 0

            # Keep around metrics for each task we watch
            with watch_task_lock:
                watch_task_info[task_id] = dict(watcher.info)

            if watcher.is_done():
                return None if watcher.is_success() else watcher.get_failure()
            log_f("Task state: " + koji.TASK_STATES[watcher.info['state']])
        except:
            except_count += 1
            # possible for watcher.update() to except during connection issue, try again
            log_f('watcher.update() exception. Trying again in 60s.\n{}'.format(traceback.format_exc()))
            if except_count >= 10:
                log_f('watcher.update() excepted 10 times. Giving up.')
                error = traceback.format_exc()
                break

        if terminate_event.wait(timeout=3 * 60):
            error = 'Interrupted'
        elif time.time() > end:
            error = 'Timeout building image'

    log_f(error + ", canceling build")
    subprocess.check_call(("brew", "cancel", str(task_id)))
    return error


def get_brew_build(nvr, product_version='', session=None):
    """5.2.2.1. GET /api/v1/build/{id_or_nvr}

    Get Brew build details.

    https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-get-apiv1buildid_or_nvr

    :param str nvr: A name-version-release string of a brew rpm/image build
    :param str product_version: The product version tag as given to ET
    when attaching a build
    :param requests.Session session: A python-requests Session object,
    used for for connection pooling. Providing `session` object can
    yield a significant reduction in total query time when looking up
    many builds.

    http://docs.python-requests.org/en/master/user/advanced/#session-objects

    :return: An initialized Build object with the build details
    :raises exceptions.BrewBuildException: When build not found

    """
    if session is not None:
        res = session.get(constants.errata_get_build_url.format(id=nvr),
                          auth=HTTPKerberosAuth())
    else:
        res = requests.get(constants.errata_get_build_url.format(id=nvr),
                           verify=ssl.get_default_verify_paths().openssl_cafile,
                           auth=HTTPKerberosAuth())
    if res.status_code == 200:
        return Build(nvr=nvr, body=res.json(), product_version=product_version)
    else:
        raise exceptions.BrewBuildException("{build}: {msg}".format(
            build=nvr,
            msg=res.text))


def find_unshipped_build_candidates(base_tag, product_version, kind='rpm'):
    """Find builds for a product and return a list of the builds only
    labeled with the -candidate tag that aren't attached to any open
    advisory.

    :param str base_tag: The tag to search for shipped/candidate
    builds. This is combined with '-candidate' to return the build
    difference.
    :param str product_version: The product version tag as given to ET
    when attaching a build
    :param str kind: Search for RPM builds by default. 'image' is also
    acceptable

    For example, if `base_tag` is 'rhaos-3.7-rhel7' then this will
    look for two sets of tagged builds:

    (1) 'rhaos-3.7-rhel7'
    (2) 'rhaos-3.7-rhel7-candidate'

    :return: A set of build strings where each build is only tagged as
    a -candidate build
    """
    if kind == 'rpm':
        candidate_builds = BrewTaggedRPMBuilds(base_tag + "-candidate")
        shipped_builds = BrewTaggedRPMBuilds(base_tag)
    elif kind == 'image':
        candidate_builds = BrewTaggedImageBuilds(base_tag + "-candidate")
        shipped_builds = BrewTaggedImageBuilds(base_tag)

    # Multiprocessing may seem overkill, but these queries can take
    # longer than you'd like
    pool = ThreadPool(cpu_count())
    results = pool.map(
        lambda builds: builds.refresh(),
        [candidate_builds, shipped_builds])
    # Wait for results
    pool.close()
    pool.join()

    # Builds only tagged with -candidate (not shipped yet)
    return candidate_builds.builds.difference(shipped_builds.builds)


def get_brew_buildinfo(build):
    """Get the buildinfo of a brew build from brew.

    :param str|int build: A build NVR or numeric ID

    :return dict buildinfo: A dict of the build info. Certain fields
    may be transformed such as 'Tags' turned into a list, and 'Extra'
    into a proper dict object instead of a JSON string

    :return str buildinfo: The raw unparsed buildinfo as a string

Note: This is different from get_brew_build in that this function
queries brew directly using the 'brew buildinfo' command. Whereas,
get_brew_build queries the Errata Tool API for other information.

This function will give information not provided by ET: build tags,
finished date, built by, etc.
    """
    query_string = "brew buildinfo {nvr}".format(nvr=build.nvr)
    rc, stdout, stderr = exectools.cmd_gather(shlex.split(query_string))
    buildinfo = {}

    if as_string and rc == 0:
        return stdout
    else:
        buildinfo = {}

        # These keys indicate content which can be whitespace split into
        # an array, ex: the value of the "Tags" key comes as a string, but
        # we can split it on space characters and have a list of tags
        # instead.
        split_keys = ['Tags']
        # The value of these keys contain json strings which we can
        # attempt to load into a datastructure
        json_keys = ['Extra']
        for line in stdout.splitlines():
            key, token, rest = line.partition(': ')
            if key in split_keys:
                buildinfo[key] = rest.split(' ')
                continue
            elif key in json_keys:
                try:
                    # Why would we use the ast module for this? Is
                    # json.loads not enough? Actually, no, it isn't
                    # enough. The <airquotes>JSON</airquotes> returned
                    # from 'brew buildinfo' uses single-quotes to wrap
                    # strings, that is not valid json. Whereas,
                    # ast.literal_eval() can handle that and load the
                    # string into a structure we can work with
                    buildinfo[key] = ast.literal_eval(rest)
                except Exception as e:
                    buildinfo[key] = rest
                    continue
            else:
                buildinfo[key] = rest
        return buildinfo


def get_tagged_image_builds(tag, latest=True):
    """Wrapper around shelling out to run 'brew list-tagged' for a given tag.

    :param str tag: The tag to list builds from
    :param bool latest: Only show the single latest build of a package
    """
    if latest:
        latest_option = '--latest'
    else:
        latest_option = ''

    query_string = "brew list-tagged {tag} {latest} --type=image --quiet".format(tag=tag, latest=latest_option)
    # --latest - Only the last build for that package
    # --type=image - Only show container images builds
    # --quiet - Omit field headers in output

    return exectools.cmd_gather(shlex.split(query_string))


def get_tagged_rpm_builds(tag, arch='src', latest=True):
    """Wrapper around shelling out to run 'brew list-tagged' for a given tag.

    :param str tag: The tag to list builds from
    :param str arch: Filter results to only this architecture
    :param bool latest: Only show the single latest build of a package
    """
    if latest is True:
        latest_flag = "--latest"
    else:
        latest_flag = ""

    query_string = "brew list-tagged {tag} {latest} --rpm --quiet --arch {arch}".format(tag=tag, latest=latest_flag, arch=arch)
    # --latest - Only the last build for that package
    # --rpm - Only show RPM builds
    # --quiet - Omit field headers in output
    # --arch {arch} - Only show builds of this architecture

    return exectools.cmd_gather(shlex.split(query_string))

# ============================================================================
# Brew object interaction models
# ============================================================================


# lifted from https://github.com/rpm-software-management/yum/blob/master/rpmUtils/miscutils.py
# Usually available in rpmUtils when yum installed, but not available always on
# newer Fedora. So, for testing, extracted to here
def splitRPMFilename(filename):
    """
    Pass in a standard style rpm fullname

    Return a name, version, release, epoch, arch, e.g.::
        foo-1.0-1.i386.rpm returns foo, 1.0, 1, i386
        1:bar-9-123a.ia64.rpm returns bar, 9, 123a, 1, ia64
    """

    if filename[-4:] == '.rpm':
        filename = filename[:-4]

    archIndex = filename.rfind('.')
    arch = filename[archIndex + 1:]

    relIndex = filename[:archIndex].rfind('-')
    rel = filename[relIndex + 1:archIndex]

    verIndex = filename[:relIndex].rfind('-')
    ver = filename[verIndex + 1:relIndex]

    epochIndex = filename.find(':')
    if epochIndex == -1:
        epoch = ''
    else:
        epoch = filename[:epochIndex]

    name = filename[epochIndex + 1:verIndex]
    return name, ver, rel, epoch, arch


def get_tagged_rpm_names(branch, arch='x86_64'):
    query_string = "brew list-tagged --inherit --latest --rpms --arch {arch} {branch}-container-build"
    # --latest - Only the last build for that package
    # --rpm - Only show RPM builds
    # --quiet - Omit field headers in output
    # --arch {arch} - Only show builds of this architecture

    rpms = []

    arches = ['noarch', arch]  # always gather noarch
    for a in arches:
        rc, stdout, stderr = exectools.cmd_gather(shlex.split(query_string.format(branch=branch, arch=a)))
        if rc == 0:
            rpms.extend(stdout.splitlines())
        else:
            raise ValueError(stderr)

    result = set([splitRPMFilename(line.strip())[0] for line in rpms])

    return result


def check_rpm_buildroot(name, branch, arch='x86_64'):
    """
    Query the buildroot used by ODCS to determine if a given RPM name
    is provided by ODCS for the given arch.
    :param str name: RPM name
    :param str branch: Current building branch, such as rhaos-3.10-rhel-7
    :param str arch: CPU architecture to search
    """
    args = locals()
    query = 'repoquery --repofrompath foo,"http://download-node-02.eng.bos.redhat.com/brewroot/repos/{branch}-ppc64le-container-build/latest/{arch}" --repoid=foo --arch {arch},noarch --whatprovides {name}'
    rc, stdout, stderr = exectools.cmd_gather(query.format(**args))
    if rc == 0:
        result = []
        stdout = stdout.strip()
        for rpm in stdout.strip().splitlines():
            n = rpm.split(':')[0]
            n = '-'.join(n.split('-')[0:-1])
            result.append(n)

        return result
    else:
        raise ValueError(stderr)


class BrewTaggedImageBuilds(object):
    """
    Abstraction around working with lists of brew tagged image
    builds. Ensures the result set is formatted correctly for this
    build type.
    """
    def __init__(self, tag):
        self.tag = tag
        self.builds = set([])

    def refresh(self):
        """Refresh or build initial list of brew builds

        :return: True if builds could be found for the given tag

        :raises: Exception if there is an error looking up builds
        """
        rc, stdout, stderr = get_tagged_image_builds(self.tag)

        print("Refreshing for tag: {tag}".format(tag=self.tag))

        if rc != 0:
            raise exceptions.BrewBuildException("Failed to get brew builds for tag: {tag} - {err}".format(tag=self.tag, err=stderr))
        else:
            builds = set(stdout.splitlines())
            for b in builds:
                self.builds.add(b.split()[0])

        return True


class BrewTaggedRPMBuilds(object):
    """
    Abstraction around working with lists of brew tagged rpm
    builds. Ensures the result set is formatted correctly for this
    build type.
    """
    def __init__(self, tag):
        self.tag = tag
        self.builds = set([])

    def refresh(self):
        """Refresh or build initial list of brew builds

        :return: True if builds could be found for the given tag

        :raises: Exception if there is an error looking up builds
        """
        rc, stdout, stderr = get_tagged_rpm_builds(self.tag)

        print("Refreshing for tag: {tag}".format(tag=self.tag))

        if rc != 0:
            raise exceptions.BrewBuildException("Failed to get brew builds for tag: {tag} - {err}".format(tag=self.tag, err=stderr))
        else:
            builds = set(stdout.splitlines())
            for b in builds:
                # The results come back with the build arch (.src)
                # appended. Remove that if it is in the string.
                try:
                    self.builds.add(b[:b.index('.src')])
                except ValueError:
                    # Raised if the given substring is not found
                    self.builds.add(b)

        return True


class Build(object):
    """An existing brew build

How might you use this object? Great question. I'd start by fetching
the details of a known build from the Errata Tool using the
/api/v1/build/{id_or_nvr} API endpoint. Then take that build NVR or ID
and the build object from the API and initialize a new Build object
from those.

Save yourself some time and use the brew.get_brew_build()
function. Give it an NVR or a build ID and it will give you an
initialized Build object (provided the build exists).

    """
    def __init__(self, nvr=None, body={}, product_version=''):
        """Model for a brew build.

        :param str nvr: Name-Version-Release (or build ID) of a brew build

        :param dict body: An object as one gets from the errata tool
        /api/v1/build/{id_or_nvr} REST endpoint. See also:
        get_brew_build() (above)

        :param str product_version: The tag (from Errata Tool) of the
        product this build will be attached to, for example:
        "RHEL-7-OSE-3.9". This is only useful when representing this
        object as an item that would be given to the Errata Tool API
        add_builds endpoint (see: Build.to_json()).
        """
        self.nvr = nvr
        self.body = body
        self.all_errata = []
        self.kind = ''
        self.path = ''
        self.attached_erratum_ids = set([])
        self.attached_closed_erratum_ids = set([])
        self.product_version = product_version
        self.buildinfo = {}
        self.process()

    def __str__(self):
        return self.nvr

    def __repr__(self):
        return "Build({nvr})".format(nvr=self.nvr)

    # Set addition
    def __eq__(self, other):
        return self.nvr == other.nvr

    # Set addition
    def __ne__(self, other):
        return self.nvr != other.nvr

    # List sorting
    def __gt__(self, other):
        return self.nvr > other.nvr

    # List sorting
    def __lt__(self, other):
        return self.nvr < other.nvr

    @property
    def open_erratum(self):
        """Any open erratum this build is attached to"""
        return [e for e in self.all_errata if e['status'] in constants.errata_active_advisory_labels]

    @property
    def attached_to_open_erratum(self):
        """Attached to any open erratum"""
        return len(self.open_erratum) > 0

    @property
    def closed_erratum(self):
        """Any closed erratum this build is attached to"""
        return [e for e in self.all_errata if e['status'] in constants.errata_inactive_advisory_labels]

    @property
    def attached_to_closed_erratum(self):
        """Attached to any closed erratum"""
        return len(self.closed_erratum) > 0

    @property
    def attached(self):
        """Attached to ANY erratum (open or closed)"""
        return len(self.all_errata) > 0

    def process(self):
        """Generate some easy to access attributes about this build so we
don't have to do extra manipulation later back in the view"""
        # Has this build been attached to any erratum?
        self.all_errata = self.body.get('all_errata', [])

        # What kind of build is this?
        if 'files' in self.body:
            # All of the files are provided. What we're trying to do
            # is figure out if this build classifies as one of the
            # kind of builds we work with: RPM builds and Container
            # Image builds.
            #
            # We decide opportunistically, hence the abrupt
            # breaks. This decision process may require tweaking in
            # the future.
            #
            # I've only ever seen OSE image builds having 1 item (a
            # tar file) in the 'files' list. On the other hand, I have
            # seen some other general product builds that have both
            # tars and rpms (and assorted other file types), and I've
            # seen pure RPM builds with srpms and rpms...
            for f in self.body['files']:
                if f['type'] == 'rpm':
                    self.kind = 'rpm'
                    self.file_type = 'rpm'
                    break
                elif f['type'] == 'tar':
                    self.kind = 'image'
                    self.file_type = 'tar'
                    break

    def add_buildinfo(self, verbose=False):
        """Add buildinfo from upstream brew"""
        date_format = '%a, %d %b %Y %H:%M:%S %Z'
        if verbose:
            click.secho('.', nl=False)
        self.buildinfo = get_brew_buildinfo(self)
        self.finished = datetime.datetime.strptime(self.buildinfo['Finished'], date_format)

    def to_json(self):
        """Method for adding this build to advisory via the Errata Tool
API. This is the body content of the erratum add_builds endpoint."""
        return {
            'product_version': self.product_version,
            'build': self.nvr,
            'file_types': [self.file_type],
        }



"""

Test the brew related functions/classes

"""

import mock
from contextlib import nested

import logging
import StringIO

import platform
(major, minor, patch) = platform.python_version_tuple()
if int(major) == 2 and int(minor) < 7:
    import unittest2 as unittest
else:
    import unittest

import exceptions
import constants
import brew
import test_structures


class TestBrew(unittest.TestCase):

    def setUp(self):
        """
        Define and provide mock logging for test/response
        """
        self.stream = StringIO.StringIO()
        logging.basicConfig(level=logging.DEBUG, stream=self.stream)
        self.logger = logging.getLogger()

    def tearDown(self):
        """
        Reset logging for each test.
        """
        logging.shutdown()
        reload(logging)

    def test_build_attached_to_open_erratum(self):
        """We can tell if a build is attached to any open erratum"""
        # Create Erratum(), create Build() using dict with all_errata
        # containing an object with 'id' matching Erratum.advisory_id
        b = brew.Build(nvr='template-service-broker-docker-v3.7.36-2',
                       body=test_structures.image_build_attached_open_json,
                       product_version='rhaos-test-7')
        self.assertTrue(b.attached_to_open_erratum)

    def test_build_attached_to_closed_erratum(self):
        """We can tell if a build is attached to any closed erratum"""
        # Use filter #1991: (Active; Product: RHOSE; sorted by newest)
        b = brew.Build(nvr='template-service-broker-docker-v3.7.36-2',
                       body=test_structures.image_build_attached_closed_json,
                       product_version='rhaos-test-7')
        self.assertTrue(b.attached_to_closed_erratum)

    def test_good_attached_brew_image_build(self):
        """We can create and process an attached image Build object"""
        b = brew.Build(nvr='template-service-broker-docker-v3.7.36-2',
                       body=test_structures.image_build_attached_json,
                       product_version='rhaos-test-7')

        self.assertEqual('template-service-broker-docker-v3.7.36-2', b.nvr)
        self.assertEqual('image', b.kind)
        self.assertEqual('tar', b.file_type)
        self.assertTrue(b.attached)

    def test_good_unattached_brew_image_build(self):
        """We can create and process an unattached image Build object"""
        b = brew.Build(nvr='cri-o-docker-v3.7.37-1',
                       body=test_structures.image_build_unattached_json,
                       product_version='rhaos-test-7')

        self.assertEqual('cri-o-docker-v3.7.37-1', b.nvr)
        self.assertEqual('image', b.kind)
        self.assertEqual('tar', b.file_type)
        self.assertFalse(b.attached)

    def test_good_attached_brew_rpm_build(self):
        """We can create and process an attached rpm Build object"""
        b = brew.Build(nvr='coreutils-8.22-21.el7',
                       body=test_structures.rpm_build_attached_json,
                       product_version='rhaos-test-7')

        self.assertEqual('coreutils-8.22-21.el7', b.nvr)
        self.assertEqual('rpm', b.kind)
        self.assertEqual('rpm', b.file_type)
        self.assertTrue(b.attached)

    def test_good_unattached_brew_rpm_build(self):
        """We can create and process an unattached rpm Build object"""
        b = brew.Build(nvr='ansible-service-broker-1.0.21-1.el7',
                       body=test_structures.rpm_build_unattached_json,
                       product_version='rhaos-test-7')

        self.assertEqual('ansible-service-broker-1.0.21-1.el7', b.nvr)
        self.assertEqual('rpm', b.kind)
        self.assertEqual('rpm', b.file_type)
        self.assertFalse(b.attached)

    def test_build_sorting(self):
        """Ensure we can sort a list of builds"""
        b1 = brew.Build(nvr='abcd-1.0.0')
        b2 = brew.Build(nvr='zyxw-1.0.0')
        # Same one as before for equality
        b3 = brew.Build(nvr='zyxw-1.0.0')

        self.assertGreater(b2, b1)
        self.assertLess(b1, b2)
        self.assertEqual(b2, b3)

    def test_build_display(self):
        """Verify brew Builds display correctly"""
        nvr = 'megafrobber-1.3.3-7'
        b = brew.Build(nvr=nvr)
        self.assertEqual(nvr, str(b))
        self.assertEqual("Build({nvr})".format(nvr=nvr), repr(b))

    def test_build_equality(self):
        """Ensure brew Builds are unique and can be tested for equality"""
        b1 = brew.Build(nvr='megafrobber-1.3.3-7')
        b2 = brew.Build(nvr='tuxracer-42')

        builds = set([])
        builds.add(b1)
        builds.add(b1)

        self.assertEqual(1, len(builds))
        self.assertTrue(b1 != b2)

    def test_rpm_build_json_formatting(self):
        """Ensure a brew Build returns proper JSON for API posting"""
        nvr = 'coreutils-8.22-21.el7'
        pv = 'rhaos-test-7'

        b = brew.Build(nvr=nvr,
                       body=test_structures.rpm_build_attached_json,
                       product_version=pv)

        expected_json = {
            'product_version': pv,
            'build': nvr,
            'file_types': ['rpm']
        }

        self.assertEqual(expected_json, b.to_json())

    def test_image_build_json_formatting(self):
        """Ensure a brew Build returns proper JSON for API posting"""
        nvr = 'template-service-broker-docker-v3.7.36-2'
        pv = 'rhaos-test-7'

        b = brew.Build(nvr=nvr,
                       body=test_structures.image_build_attached_json,
                       product_version=pv)

        expected_json = {
            'product_version': pv,
            'build': nvr,
            'file_types': ['tar']
        }

        self.assertEqual(expected_json, b.to_json())

    def test_get_brew_build_success(self):
        """Ensure a 'proper' brew build returns a Build object"""
        with nested(
                mock.patch('brew.requests.get'),
                # Mock the HTTPKerberosAuth object in the module
                mock.patch('brew.HTTPKerberosAuth')) as (get, kerb):
            nvr = 'coreutils-8.22-21.el7'
            pv = 'rhaos-test-7'
            response = mock.MagicMock(status_code=200)
            response.json.return_value = test_structures.rpm_build_attached_json
            get.return_value = response

            b = brew.get_brew_build(nvr, product_version=pv)

            # Basic object validation to ensure that the example build
            # object we return from our get() mock matches the
            # returned Build object
            self.assertEqual(nvr, b.nvr)

            get.assert_called_once_with(
                constants.errata_get_build_url.format(id=nvr),
                auth=kerb()
            )

    def test_get_brew_build_success_session(self):
        """Ensure a provided requests session is used when getting brew builds"""
        with mock.patch('brew.HTTPKerberosAuth') as kerb:
            nvr = 'coreutils-8.22-21.el7'
            pv = 'rhaos-test-7'
            # This is the return result from the session.get() call
            response = mock.MagicMock(status_code=200)
            # We'll call the json method on the result to retrieve the
            # response body from ET
            response.json.return_value = test_structures.rpm_build_attached_json
            # We create a session mock HERE to pass in when we call
            # the get_brew_build function
            session = mock.MagicMock()
            # In get_brew_build the get is method is called on the
            # session mock. We want to return our response as if we
            # actually queried the API
            session.get.return_value = response

            # Ensure we pass in the session mock we created here
            b = brew.get_brew_build(nvr, product_version=pv, session=session)

            # Basic object validation to ensure that the example build
            # object we return from our get() mock matches the
            # returned Build object
            self.assertEqual(nvr, b.nvr)

            # Our session object+get method were used, not the
            # requests.get (default) method
            session.get.assert_called_once_with(
                constants.errata_get_build_url.format(id=nvr),
                auth=kerb()
            )

    def test_get_brew_build_failure(self):
        """Ensure we notice invalid get-build responses from the API"""
        with nested(
                mock.patch('brew.requests.get'),
                # Mock the HTTPKerberosAuth object in the module
                mock.patch('brew.HTTPKerberosAuth')) as (get, kerb):
            nvr = 'coreutils-8.22-21.el7'
            pv = 'rhaos-test-7'
            # Engage the failure logic branch, will raise
            response = mock.MagicMock(status_code=404)
            response.json.return_value = test_structures.rpm_build_attached_json
            get.return_value = response

            # The 404 status code will send us down the exception
            # branch of code
            with self.assertRaises(exceptions.BrewBuildException):
                brew.get_brew_build(nvr, product_version=pv)

            get.assert_called_once_with(
                constants.errata_get_build_url.format(id=nvr),
                auth=kerb()
            )

    def test_get_tagged_image_builds_success(self):
        """Ensure the brew list-tagged command is correct for images"""
        # Any value will work for this. Let's use a real one though to
        # maintain our sanity. This matches with the example data in
        # test_structures
        tag = 'rhaos-3.9-rhel-7-candidate'
        # Big multi-line string with brew image build ouput.
        image_builds_mock_return = test_structures.brew_list_tagged_3_9_image_builds
        image_builds_mock_length = len(image_builds_mock_return.splitlines())

        with mock.patch('exectools.cmd_gather') as gexec:
            # gather_exec => (rc, stdout, stderr)
            gexec.return_value = tuple([0, image_builds_mock_return, ""])

            # Now we can test the BrewTaggedImageBuilds
            # collecter/parser class as well as the
            # get_tagged_image_builds function
            tagged_image_builds = brew.BrewTaggedImageBuilds(tag)
            # This invokes get_tagged_image_builds, which in turn
            # invokes our mocked gather_exec
            images_refreshed = tagged_image_builds.refresh()

            # Refreshing returns True after parsing the results from
            # the brew CLI command, errors will raise an exception
            self.assertTrue(images_refreshed)

            # Our example data has 59 valid parseable images listed
            self.assertEqual(image_builds_mock_length, len(tagged_image_builds.builds))

            gexec.assert_called_once_with(
                # shlex must split our command string into a list
                ['brew', 'list-tagged', 'rhaos-3.9-rhel-7-candidate', '--latest', '--type=image', '--quiet'],
                logger=self.logger
            )

    def test_get_tagged_image_builds_failed(self):
        """Ensure the brew list-tagged explodes if the brew subprocess fails"""
        # Any value will work for this. Let's use a real one though to
        # maintain our sanity. This matches with the example data in
        # test_structures
        tag = 'rhaos-3.9-rhel-7-candidate'
        # Big multi-line string with brew image build ouput.
        image_builds_mock_return = test_structures.brew_list_tagged_3_9_image_builds

        with mock.patch('exectools.cmd_gather') as gexec:
            # gather_exec => (rc, stdout, stderr)
            #
            # The '1' in position 0 is the brew subprocess return
            # code, this invokes the error raising branch of code
            gexec.return_value = tuple([1, image_builds_mock_return, ""])
            tagged_image_builds = brew.BrewTaggedImageBuilds(tag)

            with self.assertRaises(exceptions.BrewBuildException):
                tagged_image_builds.refresh()

            gexec.assert_called_once_with(
                # shlex must split our command string into a list
                ['brew', 'list-tagged', 'rhaos-3.9-rhel-7-candidate', '--latest', '--type=image', '--quiet'],
                logger=self.logger
            )

    def test_get_tagged_rpm_builds_success(self):
        """Ensure the brew list-tagged command is correct for rpms"""
        # Any value will work for this. Let's use a real one though to
        # maintain our sanity. This matches with the example data in
        # test_structures
        tag = 'rhaos-3.9-rhel-7-candidate'
        # Big multi-line string with brew rpm build ouput.
        rpm_builds_mock_return = test_structures.brew_list_tagged_3_9_rpm_builds
        rpm_builds_mock_length = len(rpm_builds_mock_return.splitlines())

        with mock.patch('exectools.cmd_gather') as gexec:
            # gather_exec => (rc, stdout, stderr)
            gexec.return_value = tuple([0, rpm_builds_mock_return, ""])

            # Now we can test the BrewTaggedRPMBuilds
            # collecter/parser class as well as the
            # get_tagged_rpm_builds function
            tagged_rpm_builds = brew.BrewTaggedRPMBuilds(tag)
            # This invokes get_tagged_rpm_builds, which in turn
            # invokes our mocked gather_exec
            rpms_refreshed = tagged_rpm_builds.refresh()

            # Refreshing returns True after parsing the results from
            # the brew CLI command, errors will raise an exception
            self.assertTrue(rpms_refreshed)

            # Our example data has 59 valid parseable rpms listed
            self.assertEqual(rpm_builds_mock_length, len(tagged_rpm_builds.builds))

            gexec.assert_called_once_with(
                # shlex must split our command string into a list
                ['brew', 'list-tagged', 'rhaos-3.9-rhel-7-candidate', '--latest', '--rpm', '--quiet', '--arch', 'src'],
                logger=self.logger
            )

    def test_get_tagged_rpm_builds_failed(self):
        """Ensure the brew list-tagged explodes if the brew subprocess fails"""
        # Any value will work for this. Let's use a real one though to
        # maintain our sanity. This matches with the example data in
        # test_structures
        tag = 'rhaos-3.9-rhel-7-candidate'
        # Big multi-line string with brew rpm build ouput.
        rpm_builds_mock_return = test_structures.brew_list_tagged_3_9_rpm_builds

        with mock.patch('exectools.cmd_gather') as gexec:
            # gather_exec => (rc, stdout, stderr)
            #
            # The '1' in position 0 is the brew subprocess return
            # code, this invokes the error raising branch of code
            gexec.return_value = tuple([1, rpm_builds_mock_return, ""])
            tagged_rpm_builds = brew.BrewTaggedRPMBuilds(tag)

            with self.assertRaises(exceptions.BrewBuildException):
                tagged_rpm_builds.refresh()

            gexec.assert_called_once_with(
                # shlex must split our command string into a list
                ['brew', 'list-tagged', 'rhaos-3.9-rhel-7-candidate', '--latest', '--rpm', '--quiet', '--arch', 'src'],
                logger=self.logger
            )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
import bugzilla

hostname = "bugzilla.redhat.com"


class TestSearchFilter(unittest.TestCase):

    def test_search_filter(self):
        """Verify the bugzilla SearchFilter works as expected"""
        field_name = "component"
        operator = "notequals"
        value = "RFE"
        expected = "&f1=component&o1=notequals&v1=RFE"

        sf = bugzilla.SearchFilter(field_name, operator, value)
        self.assertEqual(sf.tostring(1), expected)

# class TestSearchURL(unittest.TestCase):

#     def test_searchurl(self):
#         """Verify SearchURL works as expected"""
#         t = bugzilla.SearchURL()

#         t.addFilter("component", "notequals", "RFE")
#         t.addFilter("component", "notequals", "Documentation")
#         t.addFilter("component", "notequals", "Security")
#         t.addFilter("cf_verified", "notequals", "FailedQA")
#         print "url = {}".format(t)
#         assert False

if __name__ == "__main__":
    unittest.main()



"""
Test errata models/controllers
"""

import datetime
import mock
import json
from contextlib import nested

# Import the right version for your python
import platform
(major, minor, patch) = platform.python_version_tuple()
if int(major) == 2 and int(minor) < 7:
    import unittest2 as unittest
else:
    import unittest

import exceptions
import constants
import errata
import bugzilla
import brew
import test_structures

from requests_kerberos import HTTPKerberosAuth


class TestBrew(unittest.TestCase):

    def test_get_erratum_success(self):
        """Verify a 'good' erratum request is fulfilled"""
        with mock.patch('errata.requests.get') as get:
            # Create the requests.response object. The status code
            # here will change the path of execution to the not-found
            # branch of errata.get_erratum
            response = mock.MagicMock(status_code=200)
            # response's have a 'json' function that returns a dict of
            # the JSON response body ('example_erratum' defined below)
            response.json.return_value = test_structures.example_erratum
            # Set the return value of the requests.get call to the
            # response we just created
            get.return_value = response
            e = errata.get_erratum(123456)
            self.assertIsInstance(e, errata.Erratum)

    def test_get_erratum_unauthorized(self):
        """Verify an we can detect unauthorized requests"""
        with mock.patch('errata.requests.get') as get:
            # Create the requests.response object. The status code
            # here will change the path of execution to the
            # unauthorized branch of code
            response = mock.MagicMock(status_code=401)
            get.return_value = response
            with self.assertRaises(exceptions.ErrataToolUnauthenticatedException):
                errata.get_erratum(123456)

    def test_get_erratum_failure(self):
        """Verify a 'bad' erratum request returns False"""
        with mock.patch('errata.requests.get') as get:
            # Engage the not-found branch
            response = mock.MagicMock(status_code=404)
            response.json.return_value = test_structures.example_erratum
            get.return_value = response
            e = errata.get_erratum(123456)
            self.assertFalse(e)

    def test_parse_date(self):
        """Verify we can parse the date string returned from Errata Tool"""
        d_expected = '2018-03-02 15:19:08'
        d_out = datetime.datetime.strptime(test_structures.example_erratum['errata']['rhba']['created_at'], '%Y-%m-%dT%H:%M:%SZ')
        self.assertEqual(str(d_out), d_expected)

    def test_erratum_as_string(self):
        """Verify str(Erratum) is formatted correctly"""
        ds = test_structures.example_erratum
        e = errata.Erratum(body=ds)
        expected_str = "{date} {state} {synopsis} {url}".format(
            date=datetime.datetime.strptime(ds['errata']['rhba']['created_at'], '%Y-%m-%dT%H:%M:%SZ').isoformat(),
            state=ds['errata']['rhba']['status'],
            synopsis=ds['errata']['rhba']['synopsis'],
            url="{et}/advisory/{id}".format(
                et=constants.errata_url,
                id=ds['errata']['rhba']['id']))
        self.assertEqual(expected_str, str(e))

    def test_erratum_to_json(self):
        """Ensure Erratum.to_json returns the source datastructure"""
        e = errata.Erratum(body=test_structures.example_erratum)
        self.assertEqual(json.loads(e.to_json()), test_structures.example_erratum)

    def test_erratum_refresh(self):
        """Ensure Erratum.refresh does the needful"""
        with mock.patch('errata.requests.get') as get:
            # Create the requests.response object. The status code
            # here will change the path of execution to the not-found
            # branch of errata.get_erratum
            response = mock.MagicMock(status_code=200)
            # response's have a 'json' function that returns a dict of
            # the JSON response body ('example_erratum' defined below)
            response.json.return_value = test_structures.example_erratum
            # Set the return value of the requests.get call to the
            # response we just created
            get.return_value = response
            e = errata.get_erratum(123456)

            # Use the string representation for comparison, they should be identical
            original_str = str(e)

            e.refresh()
            self.assertEqual(original_str, str(e))

    def test_get_filtered_list(self):
        """Ensure we can generate an Erratum List"""
        with mock.patch('errata.requests.get') as get:
            response = mock.MagicMock(status_code=200)
            response.json.return_value = test_structures.example_erratum_filtered_list
            get.return_value = response
            res = errata.get_filtered_list()
            self.assertEqual(2, len(res))

    def test_get_filtered_list_limit(self):
        """Ensure we can generate a trimmed Erratum List"""
        with mock.patch('errata.requests.get') as get:
            response = mock.MagicMock(status_code=200)
            response.json.return_value = test_structures.example_erratum_filtered_list
            get.return_value = response
            res = errata.get_filtered_list(limit=1)
            self.assertEqual(1, len(res))

    def test_get_filtered_list_fail(self):
        """Ensure we notice invalid erratum lists"""
        with mock.patch('errata.requests.get') as get:
            response = mock.MagicMock(status_code=404)
            response.json.return_value = test_structures.example_erratum_filtered_list
            get.return_value = response
            with self.assertRaises(exceptions.ErrataToolError):
                errata.get_filtered_list()

    def test_working_erratum(self):
        """We can create an Erratum object with a known erratum from the API"""
        # If there is an error, it will raise on its own during parsing
        #
        # If the tool fails in the future due to a schema change in
        # the returned erratum object from the ET API then the
        # `example_erratum` in this test file will need to be updated.
        e = errata.Erratum(body=test_structures.example_erratum)
        self.assertEqual(type(e), type(errata.Erratum()))

    def test_add_bug(self):
        """Verify Bugs are added the right way"""
        with nested(
                mock.patch('errata.requests.post'),
                # Mock the HTTPKerberosAuth object in the module
                mock.patch('errata.HTTPKerberosAuth')) as (post, kerb):
            response = mock.MagicMock(status_code=404)
            response.json.return_value = test_structures.example_erratum_filtered_list
            post.return_value = response

            b = bugzilla.Bug(id=1337)

            # With the mocked HTTPKerberosAuth object we can now
            # create an erratum
            e = errata.Erratum(body=test_structures.example_erratum)

            # When we make the method call, we will be using the same
            # mocked ('kerb') HTTPKerberosAuth object
            e.add_bug(b)

            post.assert_called_once_with(
                constants.errata_add_bug_url.format(id=test_structures.example_erratum['content']['content']['errata_id']),
                auth=kerb(),
                json={'bug': b.id}
            )

    def test_add_builds_success(self):
        """Ensure legit builds are added correctly"""
        with nested(
                mock.patch('errata.requests.post'),
                mock.patch('errata.HTTPKerberosAuth')) as (post, kerb):
            response = mock.MagicMock(status_code=200)
            response.json.return_value = test_structures.example_erratum_filtered_list
            post.return_value = response

            pv = 'rhaos-test-7'
            e = errata.Erratum(body=test_structures.example_erratum)
            b1 = brew.Build(nvr='coreutils-8.22-21.el7',
                            body=test_structures.rpm_build_attached_json,
                            product_version=pv)
            b2 = brew.Build(nvr='ansible-service-broker-1.0.21-1.el7',
                            body=test_structures.rpm_build_unattached_json,
                            product_version=pv)
            builds = [b1, b2]

            result = e.add_builds(builds)

            # add_builds returns True on success
            self.assertTrue(result)
            # Even though we have multiple builds, the add_builds
            # endpoint allows us to make just one call, as it
            # accepts a list of builds in the request body
            self.assertEqual(post.call_count, 1)

            post.assert_called_once_with(
                constants.errata_add_builds_url.format(id=test_structures.example_erratum['content']['content']['errata_id']),
                auth=kerb(),
                json=[b1.to_json(), b2.to_json()]
            )

    def test_add_builds_failure(self):
        """Ensure failing add_builds raises correctly on a known bad status code"""
        with nested(
                mock.patch('errata.requests.post'),
                mock.patch('errata.HTTPKerberosAuth')) as (post, kerb):
            # This triggers the failure code-branch
            response = mock.MagicMock(status_code=422)
            response.json.return_value = test_structures.example_erratum_filtered_list
            post.return_value = response

            pv = 'rhaos-test-7'
            e = errata.Erratum(body=test_structures.example_erratum)
            b1 = brew.Build(nvr='coreutils-8.22-21.el7',
                            body=test_structures.rpm_build_attached_json,
                            product_version=pv)
            b2 = brew.Build(nvr='ansible-service-broker-1.0.21-1.el7',
                            body=test_structures.rpm_build_unattached_json,
                            product_version=pv)
            builds = [b1, b2]

            with self.assertRaises(exceptions.BrewBuildException):
                e.add_builds(builds)

    # Commented out until we update add_builds to handle non-422 response codes
    # def test_add_builds_failure(self):
    #     """Ensure failing add_builds raises correctly on an unknown bad status code"""
    #     with mock.patch('errata.requests.post') as post:
    #         # This triggers the failure code-branch
    #         response = mock.MagicMock(status_code=500)
    #         response.json.return_value = example_erratum_filtered_list
    #         post.return_value = response
    #         with mock.patch('errata.HTTPKerberosAuth') as kerb:
    #             pv = 'rhaos-test-7'
    #             e = errata.Erratum(body=example_erratum)
    #             b1 = brew.Build(nvr='coreutils-8.22-21.el7',
    #                             body=rpm_build_attached_json,
    #                             product_version=pv)
    #             b2 = brew.Build(nvr='ansible-service-broker-1.0.21-1.el7',
    #                             body=rpm_build_unattached_json,
    #                             product_version=pv)
    #             builds = [b1, b2]

    #             with self.assertRaises(exceptions.BrewBuildException):
    #                 result = e.add_builds(builds)


if __name__ == '__main__':
    unittest.main()



"""
Utility functions and object abstractions for general interactions
with Red Hat Bugzilla
"""

# stdlib
from subprocess import call, check_output
import urllib
import logutil

# ours
import constants

# 3rd party
import click

logger = logutil.getLogger(__name__)


def search_for_bugs(target_release, verbose=False):
    """Search the provided target_release's for bugs in the MODIFIED state

    :return: A list of Bug objects
    """
    # Example output: "target_release=3.4.z&target_release=3.5.z&target_release=3.6.z"
    target_releases_str = ''
    for release in target_release:
        target_releases_str += 'target_release={0}&'.format(release)

    # query_url = BUGZILLA_QUERY_URL.format(target_releases_str)

    query_url = SearchURL(constants.BUGZILLA_SERVER, "MODIFIED")
    query_url.addFilter("component", "notequals", "RFE")
    query_url.addFilter("component", "notequals", "Documentation")
    query_url.addFilter("component", "notequals", "Security")

    for v in constants.DEFAULT_VERSIONS:
        query_url.addVersion(v)

    for r in target_release:
        query_url.addTargetRelease(r)

    # TODO: Expose this for debugging
    if verbose:
        click.echo(query_url)

    new_bugs = check_output(
        ['bugzilla', 'query', '--ids', '--from-url="{0}"'.format(query_url)]).splitlines()

    return [Bug(id=i) for i in new_bugs]

def search_for_bug_transitions(current_state, changed_from, changed_to):
    query_url = SearchURL(constants.BUGZILLA_SERVER, current_state)
    query_url.addFilterOperator("AND_G")
    query_url.addFilter("bug_status", "changedfrom", changed_from)
    query_url.addFilter("bug_status", "changedto", changed_to)

    for v in constants.DEFAULT_VERSIONS:
        query_url.addVersion(v)

    changed_bugs = check_output(
        ['bugzilla', 'query', '--ids', '--from-url="{0}"'.format(query_url)]).splitlines()

    return [Bug(id=i) for i in changed_bugs]

class Bug(object):
    """
    Abstract interactions with bugzilla bugs
    """
    def __init__(self, id):
        """:param int id: A Bugzilla bug ID"""
        self.id = id

    def __str__(self):
        return str(self.id)

    def __repr__(self):
        return str(self)

    def add_comment(self, comment, is_private):
        """Add a comment to a bug"""
        if is_private:
            call(['bugzilla', 'modify', self.id, '--comment', comment, '--private'])
        else:
            call(['bugzilla', 'modify', self.id, '--comment', comment])

    def add_flags(self, flags=[]):  # pragma: no cover
        """Add flags to a bug"""
        for flag in flags:
            self.add_flag(flag)

    def add_flag(self, flag):
        """Add the given flag to the bug"""
        call(['bugzilla', 'modify', '--flag', '{0}+'.format(flag), self.id])

    def add_whiteboard_value(self, value):
        call(['bugzilla', 'modify', self.id, '--whiteboard', value])

    def has_whiteboard_value(self, value):
        """Check if the value is in the Whiteboard for the bug"""
        return check_output(['bugzilla', 'query', '--id', self.id, '--whiteboard', value])


class SearchFilter(object):
    """
    This represents a query filter. Each filter consists of three components:

    * field selector string
    * operator
    * field value
    """

    pattern = "&f{0}={1}&o{0}={2}&v{0}={3}"

    def __init__(self, field, operator, value):
        self.field = field
        self.operator = operator
        self.value = value

    def tostring(self, number):
        return SearchFilter.pattern.format(
            number, self.field, self.operator, self.value
        )


class SearchURL(object):

    url_format = "https://{}/buglist.cgi"

    def __init__(self, bz_host="bugzilla.redhat.com", bug_status="MODIFIED"):
        self.bz_host = bz_host
        self.bug_status = bug_status

        self.classification = "Red Hat"
        self.product = "OpenShift Container Platform"
        self.filters = []
        self.filter_operator = ""
        self.versions = []
        self.target_releases = []

    def __str__(self):
        root_string = SearchURL.url_format.format(self.bz_host)

        url = root_string + self._status_string()

        url += "&classification={}".format(urllib.quote(self.classification))
        url += "&product={}".format(urllib.quote(self.product))
        url += self.filter_operator
        url += self._filter_string()
        url += self._target_releases_string()
        url += self._version_string()

        return url

    def _status_string(self):
        return "?bug_status={}".format(self.bug_status)

    def _version_string(self):
        return "".join(["&version={}".format(i) for i in self.versions])

    def _filter_string(self):
        return "".join([f.tostring(i) for i, f in enumerate(self.filters)])

    def _target_releases_string(self):
        return "".join(["&target_release={}".format(tr) for tr in self.target_releases])

    def addFilter(self, field, operator, value):
        self.filters.append(SearchFilter(field, operator, value))

    def addFilterOperator(self, operator):
        # Valid operators:
        #   AND_G - Match ALL against the same field
        #   OR - Match separately
        self.filter_operator += "&j_top={}".format(operator)

    def addTargetRelease(self, release_string):
        self.target_releases.append(release_string)

    def addVersion(self, version):
        self.versions.append(version)
